//----------------------------------------------------------------------
// File:			ALTS.cpp
// Programmer:		David Mount
// Description:		Basic ALTS methods
// Last modified:	01/20/2016 (Version 1.1)
//----------------------------------------------------------------------
// Copyright (c) 2016 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//	Revision 1.1  01/20/2016
//		Modified search parameter defaults
//		Changed B&B to ALTS
//----------------------------------------------------------------------

#include "ALTS.h"						// ALTS definitions

using namespace std;					// make std:: accessible
using namespace alts;					// make alts:: accessible

//----------------------------------------------------------------------
// Error routine
//----------------------------------------------------------------------

void alts::error(						// error routine
	const string&	msg,				// error message
	bool			abort)				// abort afterwards
{
	if (abort) {
		cerr << "ALTS ERROR: " << msg << endl;
		cout << "ALTS ERROR: " << msg << endl;
		#ifdef WIN32					// for WIN32/Debug only
		#ifdef _DEBUG
			alts::wait();				// wait to keep output window alive
		#endif
		#endif
		exit(1);
	}
	else {
		cerr << "ALTS WARNING: " << msg << endl;
		cout << "ALTS WARNING: " << msg << endl;
	}
}

void alts::wait()						// kludge to retain output window
{
	cout << "Hit 'x' to exit program" << endl;
	char c;
	do {
		cin >> c;
	} while (c != 'x');
}

//----------------------------------------------------------------------
//	DVector methods
//----------------------------------------------------------------------

DVector::DVector(int d, Scalar x0, Scalar xx) : vector<Scalar>(d)
{
	DVectorIter it = begin();
	*(it++) = x0;
	while (it != end()) *(it++) = xx;
}

void DVector::shiftCopyTo(Scalar* v, int s) const
{
	for (int j = 0; j < getDim(); j++) v[j+s] = (*this)[j];
}

void DVector::shiftCopyFrom(Scalar* v, int s)
{
	for (int j = 0; j < getDim(); j++) (*this)[j] = v[j+s];
}

//----------------------------------------------------------------------
//	Slope methods
//----------------------------------------------------------------------

										// affine combination of slopes
Slope alts::affineComb(Scalar a, const Slope& b1, const Slope& b2)
{
	int d = b1.getDim();				// dimension
	Slope b(d);
	for (int j = 0; j < d; j++) {
		b[j] = (1-a)*b1[j] + a*b2[j];
	}
	return b;
}

//----------------------------------------------------------------------
//	VectorRange methods
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//	Plane methods
//----------------------------------------------------------------------

										// construct from intercept/slope
Plane::Plane(Scalar b0, const Slope& b) : DVector(b.getDim())
{
	PlaneIter it = begin();
	PlaneConstIter bit = b.begin() + 1;
	*(it++) = b0;
	while (it != end()) *(it++) = *(bit++);
}

										// residual with respect to p
Scalar Plane::residual(const Point& p) const
{
	int d = p.getIndep().getDim();		// dimension
	Scalar r = p.getDep();
	for (int j = 0; j < d; j++) {
		r -= (*this)[j] * p.getIndep()[j];
	}
	return r;
}

Slope Plane::getSlope() const			// extract slope part
{
	Slope s(getDim());
	for (int j = 1; j < getDim(); j++) {
		s[j] = (*this)[j];
	}
	return s;
}

//----------------------------------------------------------------------
//	PointSet methods
//----------------------------------------------------------------------

void PointSet::shiftIndepTo(			// shift-copy indeps to M[s..][t..]
	Scalar**	M,
	int			s,
	int			t) const
{
	int nRows = (int) this->size();		// number of points
	for (int i = 0; i < nRows; i++) {
		(*this)[i].getIndep().shiftCopyTo(M[i+s], t);
	}
}

void PointSet::shiftIndepFrom(			// shift-copy indeps from M[s..][t..]
	Scalar**	M,
	int			s,
	int			t)
{
	int nRows = (int) this->size();		// number of points
	for (int i = 0; i < nRows; i++) {
		(*this)[i].getIndep().shiftCopyFrom(M[i+s], t);
	}
}

void PointSet::shiftDepTo(				// shift-copy deps to v[s..]
	Scalar*		v,
	int			s) const
{
	int nRows = (int) this->size();		// number of points
	for (int i = 0; i < nRows; i++) {
		v[i+s] = (*this)[i].getDep();
	}
}

void PointSet::shiftDepFrom(			// shift-copy deps from v[s..]
	Scalar*		v,
	int			s)
{
	int nRows = (int) this->size();		// number of points
	for (int i = 0; i < nRows; i++) {
		(*this)[i].getDep() = v[i+s];
	}
}

//----------------------------------------------------------------------
//	I/O operators
//----------------------------------------------------------------------

										// output DVector
ostream& operator<<(ostream& out, const DVector& x)
{
	out.precision(ALTS_OUTPUT_PRECISION);
	out << "( ";
	DVectorConstIter it = x.begin();
	while (it != x.end()) {
		out << *(it++) << " ";
	}
	out << ")";
	return out;
}

										// output Indep
ostream& operator<<(ostream& out, const Indep& x)
{
	IndepConstIter it = x.begin();
	if (*it != 1) {
		out << "x[0] = " << *it << endl;
		error("Internal error: Indep has x[0] != 1", true);
	}
	out.precision(ALTS_OUTPUT_PRECISION);
	out << "[ ";
	it++;								// skip homogeneous coord
	while (it != x.end()) {
		out << *(it++) << " ";
	}
	out << "]";
	return out;
}

										// output Plane
ostream& operator<<(ostream& out, const Plane& b)
{
	out.precision(ALTS_OUTPUT_PRECISION);
	out << "{ ";
	PlaneConstIter it = b.begin();
	while (it != b.end()) {
		out << *(it++) << " ";
	}
	out << "}";
	return out;
}

										// output Slope
ostream& operator<<(ostream& out, const Slope& b)
{
	SlopeConstIter it = b.begin();
	if (*it != 0) {
		out << "b[0] = " << *it << endl;
		error("Internal error: Slope has b[0] != 0", true);
	}
	out.precision(ALTS_OUTPUT_PRECISION);
	out << "| ";
	it++;								// skip homogeneous coord
	while (it != b.end()) {
		out << *(it++) << " ";
	}
	out << "|";
	return out;
}

										// output Point
ostream& operator<<(ostream& out, const Point& p)
{
	out.precision(ALTS_OUTPUT_PRECISION);
	out << "< ind: " << p.getIndep() << " dep: " << p.getDep() << " >";
	return out;
}

										// output LTSResult
ostream& operator<<(ostream& out, const LTSResult& r)
{
	out.precision(ALTS_OUTPUT_PRECISION);
	out << "[ var: " << r.getVariance()
		<< " b: " << r.getPlane()
		<< " t: " << r.getTimeStamp() << " ]";
	return out;
}

										// input Indep
istream& operator>>(istream& in, Indep& x)
{
	IndepIter it = x.begin();
	*(it++) = 1;
	while (it != x.end()) {
		in >> *(it++);
	}
	return in;
}
										// input Plane
istream& operator>>(istream& in, Plane& b)
{
	IndepIter it = b.begin();
	while (it != b.end()) {
		in >> *(it++);
	}
	return in;
}
										// input Slope
istream& operator>>(istream& in, Slope& b)
{
	IndepIter it = b.begin();
	*(it++) = 0;
	while (it != b.end()) {
		in >> *(it++);
	}
	return in;
}
										// input Point
istream& operator>>(istream& in, Point& p)
{
	Scalar y;
	in >> p.getIndep() >> y;
	p.setDep(y);
	return in;
}
										// compare vector types
bool operator==(const alts::DVector& x, const alts::DVector& y)
{
	DVectorConstIter itx = x.begin();
	DVectorConstIter ity = y.begin();
	while (itx != x.end()) {
		if (*(itx++) != *(ity++)) 
			return false;
	}
	return true;
}
										// compare vector types
bool operator<=(const alts::DVector& x, const alts::DVector& y)
{
	DVectorConstIter itx = x.begin();
	DVectorConstIter ity = y.begin();
	while (itx != x.end()) {
		if (*(itx++) > *(ity++)) 
			return false;
	}
	return true;
}

//----------------------------------------------------------------------
//	Methods for LTSResult
//----------------------------------------------------------------------

										// get relative error wrt r
Scalar LTSResult::getRelErr(const LTSResult& r) const
{
	Scalar rCost = r.getCost();
	Scalar thisCost = getCost();
										// avoid 0/0
	if (rCost < SCALAR_EPS && thisCost < SCALAR_EPS) {
		return 0;
	}
	return (thisCost - rCost) / rCost;
}

										// get absolute error wrt r
Scalar LTSResult::getAbsErr(const LTSResult& r) const
{
	return getCost() - r.getCost();
}


//----------------------------------------------------------------------
//	Methods for SrchParams
//
//	This initializes a number of parameters that control the search. The
//	particular choices here are based on the parameters used in the
//	experimental section of the paper:
//
//	  A Practical Approximation Algorithm for the LTS Estimator
//	  D. M. Mount, N. Netanyahu, C. Piatko, R. Silverman, and A. Y. Wu
//	  Computational Statistics & Data Analysis
//
//	TODO: The default value for the slopeTrimQuant (the fraction of
//	sampled slopes to keep for the sake of defining the initial cell) is
//	set to a constant here. A better method is one that is described in
//	the above paper, which is to use the value c * m * (h/n)^d, where
//
//		c - a suitable constant (possibly depending on dimension)
//		m - the number of initial samples (slopeSampleSize)
//		h/n - trimming quantile
//		d - dimension (number of independent variables + 1)
//
//	To achieve the above behavior, this parameter value should be set at
//	whenever any of the above values have changed.
//----------------------------------------------------------------------

void SrchParams::setDefaults(int d)		// set default values
{
										// general parameters
	assert(d >= 1);						// dimension must be positive
	algorithm		= ADAPTIVE_LTS;		// Adaptive-LTS
	numPivots		= 2;				// number of pivots after each fit

										// parameters for Adaptive-LTS
	srchPriorList	= SrchPriorList(	// search priority (high to low)
					MAX_SAMPLES,		// cell with maximum number of samples
					MIN_LOWER_BOUND,	// cell with minimum lower bound
					MIN_UPPER_BOUND,	// cell with min upper bound
					OLDEST_CELL);		// cell that is oldest in queue
	pqOrder			= WEIGHTED_ORDER;	// use weighted order for priorities
	maxCells		= 500;				// max cells before terminating
	cellsPerLevel	= 100;				// cells per level (for diagnostics)
	slopeSampleSize	= 512; 				// no. of slope samples for init cell
	slopeExpansion	= 1; 				// expansion factor for init cell
	enableExpansion	= true;				// allow expansion
	slopeTrimQuant	= 0.25;				// trimming quantile for sample
	splitMethod		= MEDIAN_SPLIT; 	// cell splitting method
	initSlopeRange =					// initial slope range is empty
			SlopeRange(Slope(d, SCALAR_INFTY), Slope(d, -SCALAR_INFTY));
	rootSlopeRange =					// root slope range is empty
			SlopeRange(Slope(d, SCALAR_INFTY), Slope(d, -SCALAR_INFTY));

										// parameters for Fast-LTS
	numElemFits		= 500;				// number of elemental fits
	numBest			= 10;				// number of best fits to save

										// parameters for full-search-LTS
	fullSampleSize	= 0; 				// sample size for slab generation
	fullMaxSlabs	= 0; 				// max slabs to try
}

//----------------------------------------------------------------------
//	Methods for Params
//
//	Some of the defaults are set dynamically (e.g., those based on the
//	data size). We initialize these to -1 and invoke finalize() to set
//	their values. This procedure should be called after inputting all
//	the options, but before invoking the search algorithm.
//
//	The trimAbsValue is used when computing lower bounds, and the given
//	trimLowerValue is used when computing upper bounds. Even though the
//	costs are based on sets of different sizes, note that there is no
//	need to make an adjustment when comparing these quantities, because
//	they are average values, and the averaging process considers the
//	actual sizes of the sets over which they are computed.
//----------------------------------------------------------------------

void Params::setDefaults(int d)			// set default values
{
	outputLev 		= RESULTS;			// output basic results
	out				= NULL;				// output stream
	dim				= d;				// set dimension
	residAbsErr		= 0;				// no absolute residual error
	residRelErr		= 0.01;				// 1% relative residual error
	trimAbsValue	= -1;				// actual default = dataSize/2
	trimRelValue	= -1;				// actual default = 0.5
	trimAbsErr		= -1;				// actual default = 0
	trimRelErr		= -1;				// actual default = 0
	traceFreq		= 0;				// trace frequency
	doValidation	= false;			// performing validation
	Plane estimatedTrueFit = Plane(d);	// default plane
	points.clear();						// clear point set
	srchParams.setDefaults(dim);		// set defaults for search params
}

void Params::finalize() {				// finalize values
	int dataSize = (int) points.size();	// number of points
	if (dataSize < dim) {
		cerr << "Data size = " << dataSize << endl;
		error("data size must be at least as large as dimension", true);
	}
										// trimming value set both ways?
	if (trimAbsValue > 0 && trimRelValue > 0 &&
		trimAbsValue != int(floor(trimRelValue * dataSize + 0.5))) {
		error("trim-abs-value and trim-rel-value are inconsistent",
			true);
	}
	else if (trimAbsValue > 0) {		// absolute value given
		trimRelValue = double(trimAbsValue) / double(dataSize);
	}
	else if (trimRelValue > 0) {		// relative value given
		trimAbsValue = int(floor(trimRelValue * dataSize + 0.5));
	}
	else {								// neither given
		trimRelValue = 0.50;
		trimAbsValue = int(floor(trimRelValue * dataSize + 0.5));
	}

	if (trimAbsValue == 0) {			// zero trimming value?
		error("Trimming value must be positive", true);
	}
										// trimming error set both ways?
	if (trimAbsErr > 0 && trimRelErr > 0 &&
		trimAbsErr != int(floor(trimRelErr * dataSize + 0.5))) {
		error("trim-abs-err and trim-rel-err are inconsistent",
			true);
	}
	else if (trimAbsErr > 0) {			// absolute error given
		trimRelErr = double(trimAbsErr) / double(dataSize);
	}
	else if (trimRelErr > 0) {			// relative error given
		trimAbsErr = int(floor(trimRelErr * dataSize + 0.5));
	}
	else {								// neither given
		trimRelErr = 0;
		trimAbsErr = 0;
	}
										// trimming lower value
	trimLowerValue = trimAbsValue - trimAbsErr;
										// check that out given
	if (traceFreq != 0 && out == NULL) {
		error("Output stream must be specified for tracing", true);
	}
}

//----------------------------------------------------------------------
//	SubsetEnum methods
//----------------------------------------------------------------------
										// construct
SubsetEnum::SubsetEnum(int nn) : n(nn), subset(nn)
{
	reset();
}

bool SubsetEnum::reset()				// reset to empty set
{
	for (int i = 0; i < n; i++) {
		subset[i] = false;
	}
	isTrivialFlag = true;
	return (n != 0);					// have valid subsets?
}

bool SubsetEnum::next()					// advance to next
{
	bool hasMore = false;
	for (int i = n-1; i >= 0 && !hasMore; i--) {
		if (subset[i]) {				// convert trailing 1's to 0
			subset[i] = false;
		}
		else {							// convert last 0 to 1
			subset[i] = true;
			hasMore = true;
			isTrivialFlag = (i == n-1);	// check for trivial (1,1,...,1)
			for (; i >= 0 && isTrivialFlag; i--) {
				if (!subset[i]) isTrivialFlag = false;
			}
		}
	}
	return hasMore;
}

										// output current subset
ostream& operator<<(ostream& out, const SubsetEnum& s)
{
	out << "[";
	for (int i = 0; i < s.getSize(); i++) {
		out << (s[i] ? " 1" : " 0");
	}
	out << " ]";
	return out;
}

//----------------------------------------------------------------------
//	MSubsetEnum methods
//----------------------------------------------------------------------
										// construct
MSubsetEnum::MSubsetEnum(int nn, int mm) : n(nn), m(mm), subset()
{
	reset();
}

bool MSubsetEnum::reset()				// reset to start
{
	subset.clear();						// clear subset contents
	for (int i = 0; i < m; i++) {
		subset.push_back(i);
	}
	return (m != 0) && (m <= n);		// have valid subsets?
}

bool MSubsetEnum::next()				// advance to next
{
	bool hasMore = false;
	for (int i = 0; i < m && !hasMore; i++) {
		if (subset[m-i-1] < n-i-1) {
			subset[m-i-1]++;
			for (int j = m-i; j < m; j++) {
				subset[j] = subset[j-1]+1;
			}
			hasMore = true;
		}
	}
	return hasMore;
}

bool MSubsetEnum::has(int p) const		// does current set contain p?
{
	for (int i = 0; i < m; i++) {
		if (subset[i] == p) return true;
	}
	return false;
}

										// output current subset
ostream& operator<<(ostream& out, const MSubsetEnum& s)
{
	out << "{";
	for (int i = 0; i < s.getSetSize(); i++) {
		out << " " << s[i];
	}
	out << " }";
	return out;
}
