//----------------------------------------------------------------------
// File:			ALTScell.cpp
// Programmer:		David Mount
// Description:		Methods for slope cells
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#include <algorithm>					// for nth_smallest()

#include "ALTScell.h"					// slope cell definitions
#include "ALTSutilities.h"				// LTS utilities definitions
#include "ALTSnumeric.h"				// numeric utility definitions

using namespace std;					// make std:: accessible
using namespace alts;					// make alts:: accessible

//----------------------------------------------------------------------
//	Methods for SlopeCell
//----------------------------------------------------------------------

SlopeCell::SlopeCell(int d)				// default constructor
	: slopeBnds(d), upperBnd(d)
{
	cellNum = -1;
	parentNum = -1;
	level = 0;
	status = ALIVE;
	haveLower = haveUpper = false;
}

//----------------------------------------------------------------------
//	getSample - get a sample slope from a cell
//	This is currently implemented in the simplest possible manner, by
//	taking the midpoint of the cell.
//----------------------------------------------------------------------

Slope SlopeCell::getSample() const		// get sample
{
										// return cell midpoint
	if ((int) samples.size() > 0) {
		return samples.front();			// take first sample
	}
	else {								// take cell midpoint
		return affineComb(0.5, slopeBnds.lo, slopeBnds.hi);
	}
}

//----------------------------------------------------------------------
//	split - split a cell into its children cells
//
//	Split the current cell, dividing the samples among the two children.
//	The choice of splitting method is specified in the search
//	parameters. We invoke the appropriate splitting method, which
//	returns a cutting dimension and cutting value. After this, we set
//	the slope bounds for the children appropriately, and partition the
//	samples among them.
//
//	As a convenient side effect, we assign the parent cell number and
//	level to the children.
//----------------------------------------------------------------------

void SlopeCell::split(
	SlopeCell&		c1,					// low cell (modified)
	SlopeCell&		c2,					// high cell (modified)
	const Params&	params) const		// search parameters
{
	int cutDim;							// cut dimension and cut value
	Scalar cutVal;
										// determine how to split
	switch (params.srchParams.splitMethod) {

	case MIDPT_SPLIT:					// standard midpoint
		midPtSplit(params, cutDim, cutVal);
		break;

	case SL_MIDPT:						// sliding midpoint
		slMidPtSplit(params, cutDim, cutVal);
		break;

	case MEDIAN_SPLIT:					// median split
		medianSplit(params, cutDim, cutVal);
		break;

	default:
		error("Internal error in split", true);
		break;
	}

	c1.slopeBnds = slopeBnds;			// use our bounds initially
	c1.slopeBnds.hi[cutDim] = cutVal;	// adjust the cutting dimension
	c1.parentNum = cellNum;
										// level based on cellNum
	c1.level = c1.cellNum / params.srchParams.cellsPerLevel;

	c2.slopeBnds = slopeBnds;
	c2.slopeBnds.lo[cutDim] = cutVal;
	c2.parentNum = cellNum;
										// level based on cellNum
	c2.level = c2.cellNum / params.srchParams.cellsPerLevel;
										// partition the samples
	samples.partition(cutDim, cutVal, c1.samples, c2.samples);
}

//----------------------------------------------------------------------
//	midPtSplit - split through midpoint about longest dimension
//
//	We compute the longest side of the slope bounding box, and split
//	through its midpoint.
//
//	This is the fallback method, which is always possible, even when no
//	samples are available.
//----------------------------------------------------------------------

void SlopeCell::midPtSplit(
	const Params&	params,				// search parameters
	int&			cutDim,				// cutting dimension (returned)
	Scalar& 		cutVal)	const		// cutting value (returned)
{
	int d = params.dim;					// dimension
	cutDim = 0;
	Scalar maxWidth = 0;
	for (int j = 1; j < d; j++) {
		Scalar thisWidth = slopeBnds.hi[j] - slopeBnds.lo[j];
		if (thisWidth > maxWidth) {
			maxWidth = thisWidth;
			cutDim = j;					// cutDim is cell's max side
		}
	}
										// cutVal is midpt
	cutVal = (slopeBnds.lo[cutDim] + slopeBnds.hi[cutDim])/2;
}

//----------------------------------------------------------------------
//	slMidPtSplit - sliding midpoint split
//
//	We start with the midpoint split, but if all the sampled points lie
//	to one side, we slide the splitting plane in this direction until
//	hitting the first point. If there are no sample points, this
//	degenerates to midPtSplit.
//----------------------------------------------------------------------

void SlopeCell::slMidPtSplit(
	const Params&	params,				// search parameters
	int&			cutDim,				// cutting dimension (returned)
	Scalar& 		cutVal)	const		// cutting value (returned)
{
	int nSamp = (int) samples.size();	// number of samples
	if (nSamp <= 1) {					// 0 or 1 points?
		midPtSplit(params, cutDim, cutVal);	// use midpoint split
		return;
	}
	else {
										// start with midpoint split
		midPtSplit(params, cutDim, cutVal);
										// get bounding box
		SlopeRange range = samples.getBoundingBox();

		if (cutVal < range.lo[cutDim]) {// slide if necessary
			cutVal = range.lo[cutDim];
		}
		else if (cutVal > range.hi[cutDim]) {
			cutVal = range.hi[cutDim];
		}
	}
}

//----------------------------------------------------------------------
//	medianSplit - median split method
//
//	If there are are at most two samples, we defer to midpoint split. If
//	there are three or more, we compute longest side of the bounding box
//	of the point set (not the cell's bounding box), and use the median
//	value along this direction as the splitting dimension.
//----------------------------------------------------------------------

void SlopeCell::medianSplit(
	const Params&	params,				// search parameters
	int&			cutDim,				// cutting dimension (returned)
	Scalar& 		cutVal)	const		// cutting value (returned)
{
	int d = params.dim;					// dimension
	int nSamp = (int) samples.size();	// number of samples
	if (nSamp <= 1) {					// 0 or 1 points?
		midPtSplit(params, cutDim, cutVal);	// use midpoint split
	}
	else {
										// get bounding box
		SlopeRange range = samples.getBoundingBox();
		cutDim = 1;
		Scalar cutWidth = range.hi[1] - range.lo[1];
										// compute widest side
		for (int j = 2; j < d; j++) {
			Scalar thisWidth = range.hi[j] - range.lo[j];
			if (thisWidth > cutWidth) {
				cutWidth = thisWidth;
				cutDim = j;
			}
		}
		vector<Scalar> values(nSamp);	// get cutDim coordinate values
		for (int i = 0; i < nSamp; i++) {
			values[i] = samples[i][cutDim];
		}
										// select median
		int median = nSamp/2;			// index of median (0 = min))
		nth_element(values.begin(), values.begin()+median, values.end());

		cutVal = values[median];		// split at median value
	}
}

//----------------------------------------------------------------------
//	setPriority - set cell's expansion priority
//
//	Lower priority cells are expanded first. Because the priorities are
//	stored in an STL priority queue, and because the queue extracts the
//	maximum element, we negate all key values that to be minimized.
//	Those that are to be maximized (such as number of samples) are left
//	positive.
//----------------------------------------------------------------------

void SlopeCell::setPriorities(const Params& params)
{
										// list of search priorities
	const SrchPriorList& priors = params.srchParams.srchPriorList;
	SrchPriorList::const_iterator it = priors.begin();
	priorities.clear();					// remove any existing items
	int nSamples;						// (to avoid compiler warning)

	while (it != priors.end()) {		// for each priority in list
		SrchPriority thisPrior = *it;
		switch(thisPrior) {
	
		case MIN_UPPER_BOUND:			// min upper bound
			if (!haveUpper) getUpperBound(params);
										// negate for minimization
			priorities.add(-upperBnd.getVariance());
			break;
	
		case MIN_LOWER_BOUND:			// min lower bound
			if (!haveLower) getLowerBound(params);
			priorities.add(-lowerBnd);	// negate for minimization
			break;
	
		case MIN_GAP:					// min upper-lower gap
			if (!haveUpper) getUpperBound(params);
			if (!haveLower) getLowerBound(params);
										// negate for minimization
			priorities.add(-(upperBnd.getVariance()-lowerBnd));
			break;
	
		case OLDEST_CELL:				// min cell number
			priorities.add(-cellNum);
			break;
	
		case MAX_SAMPLES:				// maximum number of samples
			nSamples = (int) samples.size();
			if (nSamples <= 1)			// treat 1 sample as if 0
				nSamples = 0;
			priorities.add(nSamples);
			break;
	
		case MIN_SAMP_VARIANCE:			// min variance of samples
			priorities.add(-samples.getVariance());
			break;
	
		default:						// (shouldn't get here)
			error("Internal error in setPriority", true);
			break;
		}
		it++;
	}
}

//----------------------------------------------------------------------
//	getLowerBound - compute a cell's lower bound
//
//	The lower bound is computed by determining the maximum and minimum
//	intercepts of the lines passing through each point, subject to the
//	cell's slope bounds. It then invokes a 1-dimensional interval LTS
//	procedure on the resulting intercept intervals.
//----------------------------------------------------------------------

void SlopeCell::getLowerBound(const alts::Params& params)
{
	int d = params.dim;					// dimension
	const PointSet& pts(params.points);	// the points
	int n = (int) pts.size();			// number of points

	ScalarRangeSet intervals(n);		// storage for intercept ranges

	const Slope& slopeLo(slopeBnds.lo);	// lower and upper slope bound
	const Slope& slopeHi(slopeBnds.hi);

	for (int i = 0; i < n; i++) {		// compute intercept bounds
		Scalar bMin = pts[i].getDep();
		Scalar bMax = bMin;
		for (int j = 1; j < d; j++) {
			Scalar x = pts[i].getIndep()[j];
			if (x >= 0) {
				bMin -= x * slopeHi[j];
				bMax -= x * slopeLo[j];
			}
			else {
				bMin -= x * slopeLo[j];
				bMax -= x * slopeHi[j];
			}
		}
		intervals[i] = ScalarRange(bMin, bMax);
	}

										// solve interval LTS
	ScalarLTSResult r = intvLTS(intervals, params.trimAbsValue);

	lowerBnd = r.variance;				// save variance as lower bound
	haveLower = true;
}

//----------------------------------------------------------------------
//	getUpperBound - compute a cell's upper bound
//
//	The upper bound is computed by first sampling a representative slope
//	from the cell, and then determining the optimum y-intercept for this
//	slope by projecting the points to the y-axis along this slope (using
//	the fomula:
//
//		b[i] = y[i] - sum_j x[i][j]*b[j]
//
//	and then solving a scalar LTS problem for the resulting set of
//	intercepts. Note that we use the lower trimming bound.
//----------------------------------------------------------------------

void SlopeCell::getUpperBound(const alts::Params& params)
{
	const PointSet& pts(params.points);	// the points
	int d = params.dim;					// dimension
	int n = (int) pts.size();			// number of points

	ScalarSet intercepts(n);			// storage for intercepts

	Slope sample = getSample();			// get sample slope

	for (int i = 0; i < n; i++) {		// get intercepts for this slope
		Scalar b = pts[i].getDep();
		for (int j = 1; j < d; j++) {
			b -= pts[i].getIndep()[j] * sample[j];
		}
		intercepts[i] = b;
	}
										// solve interval LTS
	ScalarLTSResult r = scalarLTS(intercepts, params.trimLowerValue);
	Plane rb = Plane(r.loc, sample);	// result plane
										// save result and its cost
	upperBnd = LTSResult(r.variance, rb, cellNum);
										// attempt local improvement
	for (int i = 0; i < params.srchParams.numPivots; i++) {
		bool improved = upperBnd.improveFit(params.points,
									params.trimLowerValue);
		if (!improved) break;			// quit on convergence
	}
	haveUpper = true;
}

//----------------------------------------------------------------------
//	Methods for CellPriorList
//----------------------------------------------------------------------

										// lexicographical comparison
bool CellPriorList::operator<(const CellPriorList& p) const
{
	CellPriorList::const_iterator left = begin();
	CellPriorList::const_iterator right = p.begin();
										// traverse to ends
	bool result = false;
	while (!(left == end() && right == p.end())) {
		Scalar l = (left == end() ? 0 : *left);
		Scalar r = (right == p.end() ? 0 : *right);
		if (l < r) {
			result = true;				// left is smaller
			break;
		}
		if (l > r) {
			result = false;				// left is larger
			break;
		}
		if (left != end()) {
			left++;
		}
		if (right != p.end()) {
			right++;
		}
	}
	return result;						// must be equal
}

										// output priority list
std::ostream& operator<<(std::ostream& out, const alts::CellPriorList& p)
{
	out.precision(ALTS_OUTPUT_PRECISION);
	out << "( ";
	CellPriorList::const_iterator it = p.begin();
	while (it < p.end()-1) {
		out << *(it++) << " ";
	}
	out << *it << " )";
	return out;
}

//----------------------------------------------------------------------
//	trace - trace cell information
//	Print information about the current cell.
//----------------------------------------------------------------------

void SlopeCell::trace(
	const LTSResult&	best,			// best result so far
	const Params&		params)	const	// parameters
{
	ostream* out = params.out;			// output stream
										// best cost so far
	Scalar bestCost = best.getCost();
	Scalar lbCost = sqrt(lowerBnd);
	Scalar relErr = (lbCost != 0 ? (bestCost - lbCost)/lbCost : SCALAR_INFTY);
	Scalar absErr = (bestCost - lbCost);

	out->precision(ALTS_OUTPUT_PRECISION);
										// output basic info
	*out << "Cell: "		<< cellNum << "\n"
			<< "  par: "	<< std::setw(2) << parentNum
			<< " lev: "		<< std::setw(2) << level 
			<< " nsamp: "   << std::setw(3) << (int) samples.size() 
			<< " best: "	<< best.getVariance()
			<< " ( " << bestCost << " )\n"
		 << "  (lo, hi):    (" << slopeBnds.lo << ", " << slopeBnds.hi << ")\n"
		 << "  lower-bound: " << lowerBnd
			<< " (rel-err = "	<< relErr << ") "
			<< " (abs-err = "	<< absErr << ")\n";

	if (status == ALIVE) {				// live cell
		Scalar ubCost = upperBnd.getCost();
		*out << "  plane:       " << upperBnd.getPlane() << "\n"
			<< "  upper-bound: " << upperBnd.getVariance()
			<< " (error = " << (ubCost - bestCost)/bestCost << ")\n"
			<< "  priorities:  " << priorities << "\n"
			<< "  status:      alive\n";
		}
	else {								// dead cell
		*out << "  status:      dead\n";
	}
	*out << "  end-trace\n";
	out->flush();						// flush output buffer
}
