//----------------------------------------------------------------------
// File:			ALTScell.h
// Programmer:		David Mount
// Description:		Definitions for ALTS slope cells
// Last modified:	01/20/2016 (Version 1.1)
//----------------------------------------------------------------------
// Copyright (c) 2016 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//	Revision 1.1  01/20/2016
//		Changed B&B to ALTS
//----------------------------------------------------------------------

#ifndef ALTS_CELL_H
#define ALTS_CELL_H

#include "ALTS.h"						// ALTS definitions

//----------------------------------------------------------------------
//	Global types
//----------------------------------------------------------------------

namespace alts {

	enum Status {						// cell status
		ALIVE,
		DEAD };

	class CellPriorList;				// cell priority list
	class SlopeCell;					// a cell in slope space
}

//----------------------------------------------------------------------
//	CellPriorList - priority list for a cell
//
//	A priority is a sequence of scalar values. These values are in 1-1
//	correspondence with the SrchPriorList values and are compared
//	lexicographically. Natural integer values (such as number of samples
//	and cell number) are cast to Scalars. High priorities are processed
//	first, and so quantities that are to be minimized (upper bound,
//	lower bound, cell number) should be negated before storing them.
//
//	We provide two comparators, one which compares the i-th components
//	and the other compares two lists lexicographically.
//----------------------------------------------------------------------

class alts::CellPriorList : public std::vector<Scalar> {
public:
										// constructor
	CellPriorList() : std::vector<Scalar>() { }

	CellPriorList(Scalar a) : std::vector<Scalar>()
	{ push_back(a); }

	void add(Scalar a)					// add a priority
	{ push_back(a); }

										// component[i] less than
	bool isLessThan(const CellPriorList& p, int i) const
	{ return this->at(i) < p.at(i); }

										// lexicographical less than
	bool operator<(const CellPriorList& p) const;
};

										// output priority list
std::ostream& operator<<(std::ostream& out, const alts::CellPriorList& p);

//----------------------------------------------------------------------
//	Slope cell
//	A cell is an orthogonal rectangle in slope space. It is defined by a
//	two-dimensional array, containing the low and high limits of the
//	cell. As a slope object, only dimensions [1..DIM-1] are used. It
//	stores a number of quantities related to the search, the most
//	important of which are the lower bound and upper bound. The upper
//	bound stores not only the bound (as a variance) but also the
//	associated plane.
//
//	Cell public methods:
//
//		getUpperBound	Compute and return upper bound on best result.
//		getLowerBound	Compute and return lower bound on cost.
//		setPriorities	Compute and store the cell's priorities in the
//						Adaptive-LTS search. (Higher priority values
//						are expanded first by the search, so values that
//						should be minimized should be negated.) If
//						multiple priorities are given, they are
//						evaluated lexicographically.
//		getSize			Compute the cell's size, defined to be the
//						maximum of its side lengths.
//		getSample		Get a slope sample from the cell.
//		split			Split a cell into subcells.
//
//	There is a constructor that creates a SlopeCell from its ancestor in
//	the tree. In addition to updating the parent's level, we initialize
//	the priorities to that of its parent. This is needed because the
//	child cell is placed in the priorities queue prior to processing.
//	The priority values will be updated after the cell is processed.
//----------------------------------------------------------------------

class alts::SlopeCell {
public:
	int				cellNum;			// cell number
	int				parentNum;			// number of parent cell
	int				level;				// level in the ALTS search tree
	Status			status;				// cell status
	bool			haveLower;			// lower bound computed
	bool			haveUpper;			// upper bound computed
	SlopeRange		slopeBnds;			// bounds on slope
	Scalar			lowerBnd;			// lower bound on best variance
	LTSResult		upperBnd;			// upper bound on best result
	CellPriorList	priorities;			// cell priority list
	SlopeSet		samples;			// slope samples, used in splitting

private:								// private utilities

	void midPtSplit(					// midpoint split
		const Params&	params,			// search parameters
		int&			cutDim,			// cutting dimension (returned)
		Scalar& 		cutVal)	const;	// cutting value (returned)

	void slMidPtSplit(					// sliding midpoint split
		const Params&	params,			// search parameters
		int&			cutDim,			// cutting dimension (returned)
		Scalar& 		cutVal)	const;	// cutting value (returned)

	void medianSplit(					// median split
		const Params&	params,			// search parameters
		int&			cutDim,			// cutting dimension (returned)
		Scalar& 		cutVal)	const;	// cutting value (returned)

public:
	SlopeCell(int d);					// constructor given dimension

	SlopeCell(const SlopeCell& c)		// copy constructor
	{  assert(false);  }				// this should never be called!!

										// set a cell's priority
	void setPriorities(const Params& params);
										// get cell's lower bound
	void getLowerBound(const Params& params); 
										// get cell's upper bound
	void getUpperBound(const Params& params); 

	Slope getSample() const;			// get sample

	void split(							// split into subcells
		SlopeCell&		c1,				// low cell (modified)
		SlopeCell&		c2,				// high cell (modified)
		const Params&	params)			// search parameters
		const;

	void trace(							// print cell info
		const LTSResult& best,			// best result so far
		const Params&	params)	const;	// parameters
};

#endif
