//----------------------------------------------------------------------
// File:			ALTSmatrix.h
// Programmer:		David Mount
// Description:		Definitions for matrix manipulation
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#ifndef ALTS_MATRIX_H
#define ALTS_MATRIX_H

#include "ALTS.h"						// ALTS definitions

//----------------------------------------------------------------------
//	Additions to alts namespace
//----------------------------------------------------------------------

namespace alts {
	class Vect;							// a vector in point space
	class Matrix;						// a matrix in point space
										// return uniform point on sphere
	Vect uniformOnSphere(const Vect& ctr, Scalar rad);
}

//----------------------------------------------------------------------
//	Vect - a vector
//
//	A vector is a d-dimensional vector. We provide methods to cast Point
//	objects into this form, in order to eliminate the messy separation
//	between dependent and independent variables. When a point is stored
//	as a Vect, it is represented in the form (y, x1, x2, ..., x{d-1}).
//----------------------------------------------------------------------

class alts::Vect : public std::vector<Scalar> {
public:
	Vect(int d = 0, Scalar b = 0);		// construct from dim and base

	Vect(const Point& p);				// construct from Point

	int getDim() const					// get dimension
	{ return int(size()); }

	Scalar length() const;				// return length
	void normalize();					// normalize to unit length
	void complement(const Vect& u);		// complement with respect to u

	Point castToPoint() const;			// cast to point
};

//----------------------------------------------------------------------
//	Matrix - a square matrix
//
//	Matrices represent linear transformations in d-dimensional space.
//	When d matches the dimension of Point space, the matrix can be used
//	to apply transformations to points, as if they are stored in the
//	form of a Vect above, that is (y, x1, x2, ..., x{d-1}). When applied
//	to an independent vector, it is treated as the vector (0, x1, x2,
//	..., x{d-1}).
//
//	Internally a matrix is represented as an STL vector of rows, where
//	each row is an STL vector of Scalars. We assume that d is small,
//	since we often use direct copying.
//
//	The most complex method is multOrthoRand(), which pre-multiplies the
//	current matrix by a kxk random orthogonal matrix starting at index
//	i. If k = 0, then it defaults to the dimension of the matrix.
//----------------------------------------------------------------------

class alts::Matrix {
	std::vector<Vect>	M;				// the matrix [0..d-1][0..d-1]
public:
	Matrix(int d);						// construct dxd identity matrix

	int getDim() const					// get dimension
	{ return int(M.size()); }

	const Vect& operator[](int i) const	// index operator (read)
	{ return M[i]; }

	Vect& operator[](int i)				// index operator (write)
	{ return M[i]; }
										// return M * A
	Matrix operator*(const Matrix& A) const;

	void randOrthog();					// generate orthogonal rand dxd
	void shiftAndCopy(Matrix A, int s);	// shift and copy A to M[s][s]

										// apply to Plane
	Plane applyTo(const Plane& b) const;

										// apply to Vect
	Vect applyTo(const Vect& v) const;

										// apply to indep var vector
	Indep applyTo(const Indep& x) const;

										// apply to Point
	Point applyTo(const Point& p) const;
};

//----------------------------------------------------------------------
//	Output methods
//----------------------------------------------------------------------
										// output Vect
std::ostream& operator<<(std::ostream& out, const alts::Vect& v);
										// output Matrix
std::ostream& operator<<(std::ostream& out, const alts::Matrix& M);

#endif
