//----------------------------------------------------------------------
// File:			ALTSperf.h
// Programmer:		David Mount
// Description:		Definitions for performance measurement
// Last modified:	06/17/2015 (Version 1.0)
//----------------------------------------------------------------------
// Copyright (c) 2015 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//----------------------------------------------------------------------

#ifndef ALTS_PERF_H
#define ALTS_PERF_H

#include "ALTS.h"						// ALTS definitions
#include "ALTScell.h"					// ALTS slope cell definitions

//------------------------------------------------------------------------
//	Additions to alts namespace
//------------------------------------------------------------------------

namespace alts{

	class PerfCounts;					// sample count
	class PerfStat;						// sample statistics
}

//----------------------------------------------------------------------
//  PerfStat
//
//	A sample stat collects numeric (double) samples and returns some
//	simple statistics. (It is similar to BasicStat presented in
//	ALTSbasicStat.h, but it adds min and max, and as a result it cannot
//	handle removal.)
//
//		reset()			Reset to no samples
//		+= x			Include sample x
//		nSamples()		Return number of samples
//		mean()			Return mean of samples
//		variance()		Return variance
//		min()			Return minimum of samples
//		max()			Return maximum of samples
//----------------------------------------------------------------------

class alts::PerfStat {
	int n;								// number of samples
	Scalar sum;							// sum
	Scalar sumSq;						// sum of squares
	Scalar meanVal;						// mean value
	Scalar varVal;						// variance
	Scalar minVal;						// min and max
	Scalar maxVal;
	bool dirty;							// need to recompute mean & var
private:								// private utilities
	void update();						// update mean and variance
public :
	void reset()						// reset everything
	{
		n = 0;
		sum = sumSq = 0;
		minVal = SCALAR_INFTY;
		maxVal = -SCALAR_INFTY; 
		dirty = true;
	}

	PerfStat()							// constructor
	{ reset(); }

	void operator+=(Scalar x);			// add sample

	int	 nSamples() const				// number of samples
	{ return n; }

	Scalar mean();						// mean

	Scalar variance();					// variance

	Scalar stdDev()						// standard deviation
	{ return sqrt(variance()); }

	Scalar min()						// minimum
	{ return minVal; }

	Scalar max()						// maximum
	{ return maxVal; }
};

//----------------------------------------------------------------------
//  PerfCounts
//
//	A vector of performance counters. This is a STL vector of ints,
//	augmented with an operation that automatically extends the array if
//	needed.
//----------------------------------------------------------------------

class alts::PerfCounts : public std::vector<int> {
public :
	void addToLevel(int lev, int n);	// add n to level lev
};

//----------------------------------------------------------------------
//	Performance statistics
//
//	The following data and routines are used for computing performance
//	statistics.  Because these routines can slow the code down, they can
//	be activated and deactiviated by defining the PERF variable, by
//	compiling with the option: -DPERF
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//	Global counters for performance measurement
//
//	All of following statistics are computed separately for each level
//	of the search. The first group are counters that have a single value
//	for the entire level.
//
//		nProcCells			Number of cells that were processed
//		nKilledCells		Number of cells that were killed
//
//	This group are statistics that are computed for each processed cell:
//
//		lowerBnd			Lower bound on the cell's LTS variance
//		upperBnd			Upper bound on the cell's LTS variance
//----------------------------------------------------------------------

										// vector of sample statistics
typedef std::vector<alts::PerfStat> PerfStats;	

namespace alts {
										// level counters (one per level)
	extern PerfCounts nProcCells;		// number of processed cells
	extern PerfCounts nKilledCells;		// number of killed cells
}

//----------------------------------------------------------------------
//		Operation count updates
//----------------------------------------------------------------------

#ifdef PERF
#define PROC(i, n)		{nProcCells.addToLevel(i, n);}
#define KILL(i, n)		{nKilledCells.addToLevel(i, n);}
#else
#define PROC(i, n)
#define KILL(i, n)
#endif

//----------------------------------------------------------------------
//	External entry points
//----------------------------------------------------------------------

namespace alts {

	void resetLevelStats();				// reset counts/stats for each level

	void updateCellStats(				// update stats for one cell
		const SlopeCell&	c,			// the cell
		const LTSResult&	best);		// best so far

	void printStats(					// print statistics for a run
		std::ostream&		out,		// output stream
		bool				lev);		// print level statistics
}

#endif
