//----------------------------------------------------------------------
// File:			ALTStest.cpp
// Programmer:		David Mount
// Description:		Main test driver program
// Last modified:	01/20/2016 (Version 1.1)
//----------------------------------------------------------------------
// Copyright (c) 2016 University of Maryland and David Mount.  All
// Rights Reserved.
// 
// This software and related documentation is part of the Approximate
// Least-Trimmed Squares (ALTS) software.  This software is provided
// under the provisions of the GNU Public License (GPL). See the file
// ReadMe.txt for further information.
// 
// The University of Maryland (U.M.) and the authors make no
// representations about the suitability or fitness of this software for
// any purpose.  It is provided "as is" without express or implied
// warranty.
//----------------------------------------------------------------------
// History:
//	Revision 1.0  06/17/2015
//		Initial release
//	Revision 1.1  01/20/2016
//		Changed B&B to ALTS
//----------------------------------------------------------------------

#include <ctime>						// for clock()
#include <cctype>						// for isspace()
#include <cassert>						// assertion checking

#include "ALTS.h"						// basic definitions
#include "ALTSrand.h"					// random data generators
#include "ALTSnumeric.h"				// numeric utilities
#include "ALTSsearch.h"					// Adaptive-LTS search definitions
#include "ALTSfastLTS.h"				// Fast-LTS search definitions
#include "ALTSfullSearchLTS.h"			// full-search definitions

#ifdef PERF								// performance measurement only
	#include "ALTSperf.h"				// ALTS performance stats
#endif

using namespace std;					// make std:: accessible

//----------------------------------------------------------------------
// Introduction
//----------------------------------------------------------------------
// This program is the main driver for testing the approximate LTS
// program. Consider a set of n points P in d-dimensional space, where
// each point is of the form (x1, x2, ..., x{d-1}, y). We think of the
// last component, y, as the vertical direction. A (d-1)-dimensional
// hyperplane is defined by a d-element parameter vector b = (b0, b1,
// ..., b{d-1}) defined by
//
//		y = b_0 + b1*x1 + b2*x2 + ... + b{d-1}*x{d-1} = (b.x),
//
// where (b.x) denotes the dot product of b with the homogeneous vector
// representation of x as (1, x1, x2, ..., x{d-1}).  Given a hyperplane
// b, define the i-th residual of P to be
//
//		r_i(b,P) = y_i - (b.x).
//
// Let r[i](b,P) be the i-th smallest residual in terms of absolute
// value. Given a trimming parameter 1 <= m <= n and a hyperplane b,
// define LTS_m(b,P) to be LTS cost, that is, the sum of the m smallest
// squared residuals with respect to b, that is,
//
//		LTS_m(P,b) = r[1](b,P)^2 + r[2](b,P)^2 + ... + r[m](b,P)^2.
//
// Finally, LTS_m(P) is defined to be the minimum of this quantity over
// all choices of b. It is sometimes more convenient to express the
// trimming parameter as a quantile q, where 0 < q <= 1, so that m =
// ceil{q*n}.
//
// This program computes and approximation to the LTS hyperplane. There
// are a couple of ways of defining the approximation. The first is to
// approximate the residual.  We are given a real parameter residRelErr
// >= 0, and the solution is any hyperplane such that the sum of the m
// smallest squared residuals is at most (1+residRelErr)*LTS_m(P).
//
// Another natural way to define an approximation is to allow some
// flexibility in the number of residuals. We are given m' < m.  Such an
// approximation is required return a hyperplane b whose such that the
// sum of the m' smallest squared residuals with respect to b is at most
// LTS_m(P). Because fewer residuals are contributing to the sum, this
// would introduce a bias that favors the approximation algorithm. To
// correct for this, we require that the sum of the m' smallest squared
// residuals be at most (m'/m)*LTS_m(P).
//
// More generally, we may allow for both types of errors. The user
// provides two approximation parameters residRelErr and trimRelErr.
// Letting m' = ceil{m/(1+trimRelErr)}, the algorithm is required to
// produce any hyperplane b such that
//
//		LTS_m'(b,P) <= (m'/m)*(1+resErr)*LTS_m(P)
//
// As a further bit of flexibility, we allow both error parameters to be
// specified in absolute terms. The trimming error may also be expressed
// as trimAbsErr, and if so, then we use m' = m - trimAbsErr. Also,
// given a value for residAbsErr, in addition to the above condition, we
// also accept any solution b that satisfies
//
//		LTS_m'(b,P) <= (m'/m)*(LTS_m(P) + residAbsErr)
//
//----------------------------------------------------------------------
// Running this test program
//----------------------------------------------------------------------
// 
// This is a test driver for the LMS approximation. The program
// reads from the standard input and writes to the standard output.
// 
//		ALTStest [-i infile] [-o outfile]
//		
//		where:
//			infile				directive input file
//			outfile				statistics output file
//
//	If the input and/or output file are not specified, then the standard
//	input and standard output are the defaults. The program inputs
//	directives from the input file until reaching the end of file.  The
//	input file consists of a collection of directives and arguments
//	separated by white space (spaces, tabs, and new lines). Strings are
//	not quoted, and must contain any embedded white space. A "#" or "%"
//	character indicates a comment, which continues to the end of line.
//
//----------------------------------------------------------------------
// Input format
//----------------------------------------------------------------------
//
// Options:
// --------
// The program's directives are of two types, options and operations.
// Options set values, which are saved for future reference. Operations
// actually perform some action, and their behavior may depend on the
// settings of options up to that time.
//
// If an option is not specified, a default value is used. An option
// retains its value until it is set again. String inputs are not
// enclosed in quotes, and contain no embedded white space.
//
// Initialization:
// ---------------
// One of these should be called first. The dimension is one greater
// than the number of independent variables.
//
// 	initialize-with-n-indep <int>
// 								Initialize with the given number of
// 								independent variables.
// 								Default = 1
//
// 	initialize-with-n-indep <int>
// 	initialize-with-dim <int>	Initialize with the given dimension.
// 								Default = 2
//
// Options affecting output:
// -------------------------
// The following are options specify the detail in output generation.
//
//	output-level <level>		Control level of detail output where
//								<level> is one of:
//		silent					Print nothing
//		results					+= print LMS results for each trial
//		perf-stats				+= print performance stats
//		level-stats				+= print per-level performance stats
//		print-points			+= print the points
//
//								Default = "results"
//
//	trace-freq <int>			Trace frequency. Frequency f means that
//								every f cells an informative dump of the
//								current cell is printed to the output
//								file. A value of 0 means no tracing.
//								Default = 0
//
//	validate <string>			If "on", this outputs a comparison of
//								the estimated true plane of fit with the
//								result of the algorithm. If points are
//								generated randomly, then the point
//								generation method will provide the
//								estimated plane of fit. If points are
//								input from a file, the estimated plane
//								of fit can be provided with the
//								estimate-true-fit directive.
//								Default = "off"
//
//	estimate-true-fit <b0> <b1> <b2> ... <b{d-1}>
//								Estimated plane of fit. This is used in
//								conjunction with the validate option, if
//								points are input from a file.
//								Default = <0, 0, ..., 0>
//
// Size and error options:
// -----------------------
// The following options affect the data size, quantile sizes and the
// accuracy of the results computed by the approximation.  See the
// comments in the introduction above.
//
//	resid-abs-err <real>		Allowable absolute residual error.
//								Default = 0.
//
//	resid-rel-err <real>		Allowable relative residual error.
//								Default = 0.
//
//	trim-abs-value <int>		Value of m, the trimming parameter.
//								Default = 50.
//
//	trim-rel-value <real>		Value of q, the trimming quantile.
//								Default = 0.5.
//
//	trim-abs-err <int>			Absolute trimming error. A trimming
//								value of trim-abs-value - trim-abs-err
//								will be allowed in computing upper
//								bounds.
//								Default = 0.
//
//	trim-rel-err <int>			Trimming quantile error. The trimming
//								quantile will be set to trim-rel-value -
//								trim-rel-err will be used in computing
//								upper bounds.
//								Default = 0.
//
// Distributions:
// --------------
// The following distributions are used by the generatePoints to
// determine how points are generated from a source distribution.
//
// hyperplane+unif:
//		Inliers are generated uniformly along a hyperplane and outliers
//		are generated uniformly in a bounding box. The hyperplane is
//		generated by selecting the slope parameters uniformly over the
//		slope-range and an intercept uniformly over the intercept range.
//		Inliers are generated by generating the independent variable
//		from within the independent range (indep-range) and the
//		dependent range (dep-range) from the plane, and then adding a
//		Gaussian noise, given by std-dev-inlier.
//
// hyperplane+half-unif:
//		Inliers are generated uniformly along a hyperplane and outliers
//		are generated uniformly in the bounding box but only to one side
//		of the hyperplane. The only difference with this and
//		hyperplane+unif is that outliers that lie above the hyperplane
//		are filtered out.
//
//	flat+sphere:
//		Inliers are generated uniformly along a randomly rotated
//		k-dimensional flat lying in the (d-1)-dimensional space of
//		independent variables. A gaussian noise is added along the other
//		d-k-1 dimensions.  The remaining points are generated uniformly
//		on a sphere of a given radius, which is centered at the origin.
//		In addition, a fraction of the outliers, called special points,
//		are placed exactly on the equatorial hyperplane (y = 0). (This
//		is designed to be a difficult case, especially if the radius is
//		very large, and the sum of the number of inliers and special
//		points exactly equal to the trimming parameter.)
//
//	flat+rectangle:
//		The inliers are placed on an f-dimensional flat F, which itself
//		lies on a (d-1)-dimensional hyperplane H. There are two types of
//		noise added. The residStdDev noise is added relative to the
//		y-coordinate (and does not affect conditioning) and the
//		indepStdDev is added along H and does improve conditioning. The
//		outliers are uniform in a hyper-rectangle.
//
// Distribution Options:
// ---------------------
// The following options affect how random data sets are generated.
// Some options are common to different distributions. Points are
// generated from two distributions, one for the inliers, which are are
// based on the hyperplane of fit and the other for the outliers, which
// are independent of this hyperplane.
//
//	distr:type	<string>		Type of distribution, where the argument
//								is one of the above.
//
//	distr:indep-range <lo> <hi>
//								Sets bounding box enclosing all the points
//								to (lo, ..., lo) -> (hi, ..., hi)
//								Default = -1, +1
//
//	distr:dep-range <lo> <hi>	Range of dependent variable. (This
//								option is not considered for all
//								distributions.)
//								Default = (-1, +1)
//
//	distr:prob-inlier <real>	Probability of being an inlier.  The
//								argument is a real in the range [0,1]).
//								All other points are outliers.
//								Default = 1
//
//	distr:intercept-range <low> <high>
//								Range of valid intercepts (b_0).
//
//	distr:slope-range <lo> <hi>
//								Sets valid slopes b_i to
//								(lo, ..., lo) -> (hi, ..., hi)
//								Default = -1, +1
//
//	distr:resid-std-dev <real>	Residual standard deviation for the
//								inliers. The exact interpretation
//								depends on the inlier distribution.
//								Default = 0
//
//	distr:indep-std-dev <real>	Independent variable standard deviation
//								for flat+sphere and flat+rectangle.
//								Default = 0
//
//	distr:flat-dim <int>		Dimension of flat for flat+sphere and
//								flat+rectangle. Should be between 1 and
//								the number of independent variables.
//								Default = 1
//
//	distr:radius <real>			Radius used in flat+sphere.
//								Default = 1
//
//	distr:prob-special <real>	Probability of an outlier being special.
//								Default = 0
//
// Algorithm options:
// ------------------
// The following are options affect behavior of the search algorithms.
// There are three different algorithms. The main one is the geometric
// branch-and-bound algorithm, called adaptive-LTS, but we also provide
// a simplified variation of the fast-LTS heuristic of Rousseeuw and van
// Driessen and an exhaustive search algorithm. This algorithm works by
// generating a number of random elemental fits and then applying some
// number of local improvement steps. The exhaustive search algorithm,
// called full-search-LTS, generates all possible inlier subsets by
// deriving the slab determined by every subset of d+1 points. 
//
// General options:
// ------------------
//	num-pivots <int>			Number of pivot steps to use in
//								incremental improvement.
//								Default = 0
//
// Options for the Adaptive-LTS algorithm:
// -------------------------------------------
// Note that either slope-range or slope-sample-size must be specified
// and non-empty (non-zero) to determine the initial slope range for the
// search.
//
// The priority of a cell indicates the order in which it will be
// processed. There are a number of different choices for how to set the
// priorities. Note that some of these values tend to produce ties
// (e.g., min-lower-bound tends to be zero at the start of the search
// process when cells are large and max-samples tends to be zero after
// the number of cells processed exceeds the number of slope samples).

// We provide two methods for organizing the priority queue. First, the
// entries are maintained in lexicographic order, so that the first
// search priority has the highest priority. Note that oldest-cell can
// never produce a tie (because it is based on the cell number, which is
// distinct for all cells). Real-valued priorities, such as
// min-upper-bound and min-lower-bound are unlikely to produce ties,
// unless they are zero. Second, we cycle among the various search
// priorities. This latter method was designed to avoid repeatedly
// making poor selections, when the top selection criteria is not a good
// predictor.
//
//	ALTS:clear-priority			Clear the list of search priorities.
//
//	ALTS:add-search-priority <string> Add a new priority to the list:
//
//								min-upper-bound	= min upper bound first
//								min-lower-bound	= min lower bound first
//								min-gap	= min upper-lower gap first
//								min-samp-variance = min variance of samples
//								oldest-cell	= oldest cell first
//								max-samples	= cell with most samples first
//
//								Default = (min-lower-bound, oldest-cell)
//	ALTS:use-lexicographic-order	Use lexicographical heap ordering.
//
//	ALTS:use-cyclic-order		Use cyclic heap ordering.
//
//	ALTS:slope-range <lo> <hi>
//								Range of slopes to use for the original
//								cell in the adaptive-LTS search.
//
//	ALTS:slope-sample-size <int> Number of samples to take from the slope
//								Default = (empty range)
//								space in initializing the search space.
//								Default = 0
//
//	ALTS:slope-expansion <real>	Expansion factor (or contraction if < 1)
//								to be applied to the bounding box given
//								by the slope samples.
//								Default = 1
//
//	ALTS:slope-trim-quant <real>	Trimming quantile for slope bounding
//								box. This fraction of the slope samples
//								are retained in order to form the
//								bounding box for the root slope cell.
//								This is implemented as a sort of robust
//								bounding box.
//								Default = 1
//
//	ALTS:split-method <string>	Method for splitting cells. The
//								midpt-split splits at the midpoint of
//								the longest side, the median split
//								splits at the median of the longest side
//								of the tightest bound box for the
//								samples, and the sliding midpoint starts
//								the midpoint split, but if the points
//								all lie to one side, the cutting plane
//								slides until it hits a point.
//
//								midpt-split = split about cell's midpoint
//								sliding-midpt = sliding midpoint
//								median-split = split about median
//								
//								Default: midpt-split
//
//	ALTS:max-cells <int>		Terminate if this many cells have been
//								processed. If zero, then run until
//								termination.
//								Default = 0
//
//	ALTS:cells-per-level <int>	Number of cells per level. Levels are
//								used to aggregate statistics into
//								groups. Used only with performance
//								evaluation, when the output level is set
//								to level-stats or higher.
//								Default = 100
//
// Options for the Fast LTS Algorithm:
// -----------------------------------
// The following are options affect behavior of the fast-LTS algorithm.
// The default values are based on the values suggested by Rousseeuw and
// van Driessen.
//
//	FLTS:num-elem-fits <int>	Number of elemental fits.
//								Default = 500
//
//	FLTS:num-pivots <int>		Number of pivot steps to use in
//								incremental improvement.
//								Default = 0
//
//	FLTS:num-best <int>			Number of best fits to keep. We will
//								perform improvements on all of these
//								until convergence.
//								Default = 10
//
// Options for the Full Search Algorithm:
// --------------------------------------
// The following are options affect behavior of the full search
// algorithm.
//
//	Full:sample-size <int>		Number of random points to use for the
//								generation of slabs. The algorithm takes
//								O(m^{d+1} (n*log n)) time, where m is
//								this value. A value of 0 means that all
//								the data points are used.
//								Default = 0
//
//	Full:max-slabs <int>		Maximum number of slabs to try. If zero,
//								then run until termination.
//								Default = 0
//----------------------------------------------------------------------
//
// Operations:
// -----------
// The following commands actually do some action. They are evaluated at
// the moment they are input, and their action may depend on the values
// of the options set up to that time.
//
//	label <string>				Output an informative label to the
//								output file. There can be no embedded
//								white space in the string.
//	
//	clear-points				Erase the current set of points.
//
//	generate-points <int>		Generate the given number of points
//								according to the probability
//								distribution parameters and append them
//								to the current point set.
//
//	read-points <file>			Read point set from the specified file,
//								and append them to the current point
//								set.
//								Default = (none---file name required)
//
//	scramble-points				Scramble the point set. This causes the
//								current point set to be randomly
//								permuted.  In theory the LTS algorithm
//								is entirely independent of any ordering
//								within the data set, but this option
//								guarantees this. (Note that our random
//								point generators generate inliers before
//								outliers, so this is probably a good
//								idea.)
//								Default = "off".
//
//	seed <int>					Set the seed used for random numbers.
//								Setting the seed just prior to running
//								trials guarantees that all data for the
//								trial will be generated using this seed.
//
//	run-using <string>			Run, using the given method:
//
//								adaptive-lts
//								fast-lts
//								full-search-lts
//
//								Default = adaptive-lts
//----------------------------------------------------------------------

//----------------------------------------------------------------------
//	Global constants
//----------------------------------------------------------------------

										// clock tics per second
const int	ALTS_CLOCKS_PER_SEC	= 1000000;	

//----------------------------------------------------------------------
// Globals
//	The input and output streams default to the standard input and
//	output. They can be redefined in getCmdArgs, when command-line
//	arguments are processed.
//----------------------------------------------------------------------

static istream* in = &cin;				// input stream
static ostream* out = &cout;			// output stream

//----------------------------------------------------------------------
//	Global functions declarations
//----------------------------------------------------------------------

void readDirectives(int d);				// read and process directives

void readPoints(						// read points from file
	int						d,			// dimension
	alts::PointSet&			points,		// point set (modified)
	const string&			fileName);	// file name

void scramblePoints(					// randomly permute points
	alts::PointSet&			points);	// point set (modified)

void printDistrSummary(					// print summary of distribution
	const alts::DistrParams&	distrParams);

void doValidation(						// validate results
	const alts::Params&		params,		// input parameters
	const alts::LTSResult&	result);	// the result

void printSummary(						// print summary of data for run
	const alts::Params&		params,		// input parameters
	const alts::DistrParams& distrParams, // distribution parameters
	const alts::LTSResult&	result,		// the results
	double					cpuSec);	// execution time

void printPoints(
	int						dim,		// dimension
	const alts::PointSet&	points);	// input parameters

//----------------------------------------------------------------------
//	Tables of names and associated enumerated types
//
//	WARNING: Before changing any of these tables, check with ALTS.h and
//	ALTSrand.h for the names of the corresponding enumerations.
//----------------------------------------------------------------------

string outLevNames[] = {				// output levels
		"silent",						// print nothing
		"results",						// print LMS results for each trial
		"perf-stats",					// print performance stats
		"level-stats",					// print per-level performance stats
		"print-points"};				// print the points

string algoNames[] = {					// algorithms
		"adaptive-lts",					// Adaptive-LTS
		"fast-lts",						// Fast-LTS
		"full-search-lts"};				// full-search LTS

string priorityNames[] = {				// search priorities
		"min-upper-bound",				// cell with min upper bound
		"min-lower-bound",				// cell with min lower bound
		"min-gap",						// cell with min upper-lower gap
		"oldest-cell",					// cell with lowest cell number
		"max-samples",					// cell with most samples
		"min-samp-variance"};			// cell with min sample variance

string splitNames[] = {					// split methods
		"midpt-split",					// split about cell's midpoint
		"sliding-midpt",				// sliding midpoint
		"median-split"};				// split about median

string distrNames[] = {					// distributions
		"hyperplane+unif",				// plane with uniform outliers
		"hyperplane+half-unif",			// plane with one-sided outliers
		"flat+sphere",					// flat with sphere outliers
		"flat+rectangle"};				// flat with rectangle outliers

//----------------------------------------------------------------------
//	Table lookup utility
//	Returns index of search string or -1 if no found.
//----------------------------------------------------------------------

int lookUp(
	const string&	name,				// name to lookup
	string			nameList[],			// list of names
	int				size)				// number of names in list
{
	int i;
	for (i = 0; i < size; i++)
		if (name == nameList[i]) break;
	return (i >= size ? -1 : i);
}

//----------------------------------------------------------------------
// getCmdArgs - get and process command line arguments
//
//		Syntax:
//		... [-i infile] [-o outfile]
//		
//		where:
//			infile				directive input file
//			outfile				statistics output file
//
//		If file is not specified, then the standard input and standard
//		output are the defaults.
//----------------------------------------------------------------------

void getCmdArgs(int argc, char* argv[])
{
	int i = 1;
	while (i < argc) {					// read arguments
		string currentArg(argv[i]);
		if (currentArg == "-i") {		// -i option
										// create new stream
			ifstream* fin = new ifstream;
			fin->open(argv[++i], ios::in);
			if (!(*fin)) {
				alts::error("Cannot open input file", true);
			}
			in = fin;					// make this the input stream
		}
		else if (currentArg == "-o") {	// -o option
										// create new stream
			ofstream* fout = new ofstream;
			fout->open(argv[++i], ios::out);
			if (!(*fout)) {
				alts::error("Cannot open output file", true);
			}
			out = fout;					// make this the output stream
		}
		else {							// illegal syntax
			cerr << "Syntax:\n"
				 << "ALTStest [-i infile] [-o outfile]\n"
				 << "\n"
				 << "    where:\n"
				 << "    infile         directive input file\n"
				 << "    outfile        statistics output file\n";
			exit(1);					// exit
		}
		i++;
	}
}

//----------------------------------------------------------------------
// getDirective - skip comments and read next directive
//		Returns true if directive read, and false if eof seen.
//----------------------------------------------------------------------

bool skipComment(						// skip any comments
	istream*		in)					// input stream
{
	char ch = 0;
										// skip whitespace
	do { in->get(ch); } while (isspace(ch) && !in->eof());
	while (ch == '#' && !in->eof()) {	// comment?
										// skip to end of line
		do { in->get(ch); } while(ch != '\n' && !in->eof());
										// skip whitespace
		do { in->get(ch); } while(isspace(ch) && !in->eof());
	}
	if (in->eof()) return false;			// end of file
	in->putback(ch);						// put character back
	return true;
}

bool getDirective(						// read directive
	istream*		in,					// input stream
	string&		directive)				// directive storage
{
	if (!skipComment(in))				// skip comments
		return false;					// found eof along the way?
	*in >> directive;					// read directive
	return true;
}

//----------------------------------------------------------------------
// getLabel - read in a string from here to eol as a label
//----------------------------------------------------------------------

void getLabel(
	istream*			in,				// input stream
	string&				label)			// directive storage
{
	char ch;
	string temp = "";
										// skip whitespace
	do { in->get(ch); } while (isspace(ch) && !in->eof());
								
										// read to eol or comment
	while (ch != '\n' && ch != '#' && !in->eof()) {
		temp += ch;
		in->get(ch);
	}
	if (ch == '#') in->putback(ch);		// put '#' character back
	label = temp;
}

//----------------------------------------------------------------------
//	main program
//
//	The main program processes command-line arguments, processes input
//	directives, invokes the ALTS program, and prints results. We first
//	read the initialization (which gives the dimension) and then invoke
//	readDirectives() to process the rest.
//----------------------------------------------------------------------

int main(int argc, char* argv[])
{
	string				directive;		// input directive
	string				stringArg;		// a string argument
	int					dim;			// dimension

	getCmdArgs(argc, argv);				// get command-line arguments

	*out << "------------------------------------------------------------\n"
		 << "ALTStest: Version " << ALTS_VERSION << " "
		 		<< ALTS_VERSION_CMT << "\n"
		 << "    Copyright: " << ALTS_COPYRIGHT << ".\n"
		 << "    Latest Revision: " << ALTS_RELEASE_DATE << ".\n"
		 << "------------------------------------------------------------\n\n";

	getDirective(in, directive);		// input first directive

	if (directive =="initialize-with-n-indep") { 
		int nIndep;						// number of independent vars
		*in >> nIndep;
		dim = nIndep + 1;				// dimension
		if (nIndep < 1) {
			cerr << "n-indep: " << nIndep << "\n";
			alts::error("Need at least one independent variable", true);
		}
	}
	else if (directive == "initialize-with-dim") {
		*in >> dim;
		if (dim < 2) {
			cerr << "dim: " << dim << "\n";
			alts::error("Dimension must be at least 2", true);
		}
	}
	else {
		alts::error("First directive must be initialization", true);
	}

	readDirectives(dim);				// invoke main processing loop

	#ifdef WIN32						// for WIN32/Debug only
	#ifdef _DEBUG
		alts::wait();					// wait to keep output window alive
	#endif
	#endif
}

//----------------------------------------------------------------------
//	readDirectives
//	This is the main procedure for processing directives.
//----------------------------------------------------------------------

void readDirectives(int d)
{
	alts::Params		params(d);		// execution parameters
	alts::DistrParams	distrParams(d);	// distribution parameters
	string				directive;		// input directive
	string				stringArg;		// a string argument
	int					intArg;			// an integer argument

	bool dataSummarized = false;		// data not yet summarized
	bool slopeRangeGiven = false;		// slope range not yet given

	params.setDefaults(d);				// initialize parameters
	params.out = out;					// initialize output stream
	distrParams.setDefaults();			// ...and distribution parameters

	//------------------------------------------------------------------
	//	Main input loop - read and process input directive
	//------------------------------------------------------------------
	while (getDirective(in, directive)) {
		//--------------------------------------------------------------
		//	output level option
		//--------------------------------------------------------------
		if (directive == "output-level") {
			*in >> stringArg;			// input name and translate
			intArg = lookUp(stringArg, outLevNames, alts::N_OUTPUT_LEVS);
			if (intArg < 0) {			// not recognized
				cerr << "Level: " << stringArg << "\n";
				alts::error("Unknown output level option", true);
			}
										// save it
			params.outputLev = (alts::OutputLev) intArg;
		}
		//--------------------------------------------------------------
		//	trace-freq option
		//--------------------------------------------------------------
		else if (directive ==  "trace-freq") {
			*in >> params.traceFreq;
		}
		//--------------------------------------------------------------
		//	validate option
		//--------------------------------------------------------------
		else if (directive ==  "validate") {
			*in >> stringArg;			// either "on" or "off"
			if (stringArg == "off") {
				params.doValidation = false;
			}
			else if (stringArg == "on") {
				params.doValidation = true;
			}
			else {
				cerr << "Argument: " << stringArg << "\n";
				alts::error("Argument must be 'on' or 'off'", true);
			}
		}
		//--------------------------------------------------------------
		//	estimate-true-fit option
		//--------------------------------------------------------------
		else if (directive ==  "estimate-true-fit") {
			*in >> params.estimatedTrueFit;
		}
		//--------------------------------------------------------------
		//	resid-abs-err and resid-rel-err options
		//--------------------------------------------------------------
		else if (directive == "resid-abs-err") {
			*in >> params.residAbsErr;
			if (params.residAbsErr < 0) {
				alts::error("resid-abs-err must be nonnegative", true);
			}
		}
		else if (directive == "resid-rel-err") {
			*in >> params.residRelErr;
			if (params.residRelErr < 0) {
				alts::error("resid-rel-err must be nonnegative", true);
			}
		}
		//--------------------------------------------------------------
		//	trim-abs-value and trim-rel-value options
		//		Only one of these should be provided. Giving one
		//		disables the other. 
		//--------------------------------------------------------------
		else if (directive == "trim-abs-value") {
			*in >> params.trimAbsValue;
			if (params.trimAbsValue < 0) {
				alts::error("trim-abs-value must be nonnegative", true);
			}
			params.trimRelValue = -1;	// disable the other
		}
		else if (directive == "trim-rel-value") {
			*in >> params.trimRelValue;
			if (params.trimRelValue < 0) {
				alts::error("trim-rel-value must be nonnegative", true);
			}
			params.trimAbsValue = -1;	// disable the other
		}
		//--------------------------------------------------------------
		//	trim-abs-err and trim-rel-err options
		//		Only one of these should be provided. Giving one
		//		disables the other.
		//--------------------------------------------------------------
		else if (directive == "trim-abs-err") {
			*in >> params.trimAbsErr;
			if (params.trimAbsErr < 0) {
				alts::error("trim-abs-err must be nonnegative", true);
			}
			params.trimRelErr = -1;		// disable the other
		}
		else if (directive == "trim-rel-err") {
			*in >> params.trimRelErr;
			if (params.trimRelErr < 0) {
				alts::error("trim-rel-err must be nonnegative", true);
			}
			params.trimAbsErr = -1;		// disable the other
		}
		//--------------------------------------------------------------
		//	num-pivots option
		//--------------------------------------------------------------
		else if (directive == "num-pivots") {
			*in >> params.srchParams.numPivots;
		}
		//--------------------------------------------------------------
		//	ALTS:clear-search-priority option
		//--------------------------------------------------------------
		else if (directive == "ALTS:clear-search-priority") {
			params.srchParams.srchPriorList.clear();
		}
		//--------------------------------------------------------------
		//	ALTS:add-search-priority option
		//--------------------------------------------------------------
		else if (directive == "ALTS:add-search-priority") {
			*in >> stringArg;			// input name and translate
			intArg = lookUp(stringArg, priorityNames, alts::N_PRIORITIES);
			if (intArg < 0) {			// not recognized
				cerr << "ALTS:add-search-priority: " << stringArg << "\n";
				alts::error("Unknown search priority type", true);
			}
			params.srchParams.srchPriorList.add(static_cast<alts::SrchPriority> (intArg));
		}
		//--------------------------------------------------------------
		//	ALTS:use-lex-order
		//--------------------------------------------------------------
		else if (directive == "ALTS:use-lexicographic-order") {
			params.srchParams.pqOrder = alts::LEX_ORDER;
		}
		//--------------------------------------------------------------
		//	ALTS:use-cyclic-order
		//--------------------------------------------------------------
		else if (directive == "ALTS:use-cyclic-order") {
			params.srchParams.pqOrder = alts::CYCLIC_ORDER;
		}
		//--------------------------------------------------------------
		//	ALTS:use-weighted-order
		//--------------------------------------------------------------
		else if (directive == "ALTS:use-weighted-order") {
			params.srchParams.pqOrder = alts::WEIGHTED_ORDER;
		}
		//--------------------------------------------------------------
		//	ALTS:slope-range
		//--------------------------------------------------------------
		else if (directive == "ALTS:slope-range") {
			alts::Scalar slopeLo, slopeHi;
			*in >> slopeLo >> slopeHi;
			if (slopeLo > slopeHi) {
				alts::error("Low slope cannot exceed high", true);
			}
			params.srchParams.initSlopeRange.lo = alts::Slope(params.dim, slopeLo);
			params.srchParams.initSlopeRange.hi = alts::Slope(params.dim, slopeHi);
			slopeRangeGiven = true;
		}
		//--------------------------------------------------------------
		//	ALTS:slope-sample-size option
		//--------------------------------------------------------------
		else if (directive == "ALTS:slope-sample-size") {
			*in >> params.srchParams.slopeSampleSize;
		}
		//--------------------------------------------------------------
		//	ALTS:slope-expansion option
		//--------------------------------------------------------------
		else if (directive == "ALTS:slope-expansion") {
			*in >> params.srchParams.slopeExpansion;
		}
		//--------------------------------------------------------------
		//	ALTS:disable-expansion option
		//--------------------------------------------------------------
		else if (directive == "ALTS:disable-expansion") {
			params.srchParams.enableExpansion = false;
		}
		//--------------------------------------------------------------
		//	ALTS:slope-trim-quant option
		//--------------------------------------------------------------
		else if (directive == "ALTS:slope-trim-quant") {
			*in >> params.srchParams.slopeTrimQuant;
		}
		//--------------------------------------------------------------
		//	ALTS:split-method option
		//--------------------------------------------------------------
		else if (directive == "ALTS:split-method") {
			*in >> stringArg;			// input name and translate
			intArg = lookUp(stringArg, splitNames, alts::N_SPLITS);
			if (intArg < 0) {			// not recognized
				cerr << "ALTS:split-method: " << stringArg << "\n";
				alts::error("Unknown split method type", true);
			}
			params.srchParams.splitMethod = (alts::SplitMethod) intArg;
		}
		//--------------------------------------------------------------
		//	ALTS:max-cells option
		//--------------------------------------------------------------
		else if (directive == "ALTS:max-cells") {
			*in >> params.srchParams.maxCells;
		}
		//--------------------------------------------------------------
		//	ALTS:cells-per-level option
		//--------------------------------------------------------------
		else if (directive == "ALTS:cells-per-level") {
			*in >> params.srchParams.cellsPerLevel;
		}
		//--------------------------------------------------------------
		//	FLTS:num-elem-fits option
		//--------------------------------------------------------------
		else if (directive == "FLTS:num-elem-fits") {
			*in >> params.srchParams.numElemFits;
		}
		//--------------------------------------------------------------
		//	FLTS:num-best option
		//--------------------------------------------------------------
		else if (directive == "FLTS:num-best") {
			*in >> params.srchParams.numBest;
		}
		//--------------------------------------------------------------
		//	Full:sample-size option
		//--------------------------------------------------------------
		else if (directive == "Full:sample-size") {
			*in >> params.srchParams.fullSampleSize;
		}
		//--------------------------------------------------------------
		//	Full:max-slabs option
		//--------------------------------------------------------------
		else if (directive == "Full:max-slabs") {
			*in >> params.srchParams.fullMaxSlabs;
		}
		//--------------------------------------------------------------
		//	distribution option
		//--------------------------------------------------------------
		else if (directive == "distr:type") {
			*in >> stringArg;
			intArg = lookUp(stringArg, distrNames, alts::N_DISTRIBS);
			if (intArg < 0) {
				cerr << "Method =  \"" << stringArg << "\"\n";
				alts::error("Illegal distribution type", true);
			}
			distrParams.distrib = (alts::Distrib) intArg;
		}
		//--------------------------------------------------------------
		//	distribution option: distr:indep-range and distr-dep-range
		//--------------------------------------------------------------
		else if (directive == "distr:indep-range") {
			alts::Scalar rangeLo, rangeHi;
			*in >> rangeLo >> rangeHi;
			if (rangeLo > rangeHi) {
				alts::error("Low range cannot exceed high", true);
			}
			for (int j = 1; j < params.dim; j++) {
				distrParams.indepRange.lo[j] = rangeLo;
				distrParams.indepRange.hi[j] = rangeHi;
			}
		}
		else if (directive == "distr:dep-range") {
			*in >> distrParams.depRange.lo >> distrParams.depRange.hi;
			if (distrParams.depRange.lo > distrParams.depRange.hi) {
				alts::error("Low range cannot exceed high", true);
			}
		}
		//--------------------------------------------------------------
		//	distribution option: distr:prob-inlier
		//--------------------------------------------------------------
		else if (directive == "distr:prob-inlier") {
			*in >> distrParams.probInlier;
			if (distrParams.probInlier < 0 || distrParams.probInlier > 1) {
				alts::error("Inlier probability must be in [0,1]", true);
			}
		}
		//--------------------------------------------------------------
		//	distribution option: distr:intercept-range
		//--------------------------------------------------------------
		else if (directive == "distr:intercept-range") {
			*in >> distrParams.interceptRange.lo
				>> distrParams.interceptRange.hi;
			if (distrParams.interceptRange.lo >
				distrParams.interceptRange.hi) {
				alts::error("Low intercept range cannot exceed high", true);
			}
		}
		//--------------------------------------------------------------
		//	distribution option: distr:slope-range
		//--------------------------------------------------------------
		else if (directive == "distr:slope-range") {
			alts::Scalar slopeLo, slopeHi;
			*in >> slopeLo >> slopeHi;
			if (slopeLo > slopeHi) {
				alts::error("Low slope cannot exceed high", true);
			}
			distrParams.slopeRange.lo = alts::Slope(params.dim, slopeLo);
			distrParams.slopeRange.hi = alts::Slope(params.dim, slopeHi);
		}
		//--------------------------------------------------------------
		//	distribution option: resid-std-dev
		//--------------------------------------------------------------
		else if (directive == "distr:resid-std-dev") {
			*in >> distrParams.residStdDev;
			if (distrParams.residStdDev < 0) {
				alts::error("distr:resid-std-dev must be nonnegative", true);
			}
		}
		//--------------------------------------------------------------
		//	distribution option: indep-std-dev
		//--------------------------------------------------------------
		else if (directive == "distr:indep-std-dev") {
			*in >> distrParams.indepStdDev;
			if (distrParams.indepStdDev < 0) {
				alts::error("distr:indep-std-dev must be nonnegative", true);
			}
		}
		//--------------------------------------------------------------
		//	distribution option: flat-dim
		//--------------------------------------------------------------
		else if (directive == "distr:flat-dim") {
			*in >> distrParams.flatDim;
			if (distrParams.flatDim <= 0) {
				alts::error("distr:flat-dim must be positive", true);
			}
		}
		//--------------------------------------------------------------
		//	distribution option: radius
		//--------------------------------------------------------------
		else if (directive == "distr:radius") {
			*in >> distrParams.radius;
			if (distrParams.radius < 0) {
				alts::error("distr:radius must be nonnegative", true);
			}
		}
		//--------------------------------------------------------------
		//	distribution option: prob-special
		//--------------------------------------------------------------
		else if (directive == "distr:prob-special") {
			*in >> distrParams.probSpecial;
			if (distrParams.probSpecial < 0) {
				alts::error("distr:prob-special must be nonnegative", true);
			}
		}
		//--------------------------------------------------------------
		//	label operation
		//--------------------------------------------------------------
		else if (directive == "label") {
			getLabel(in, stringArg);
			cerr << "<" << stringArg << ">" << endl;
			if (params.outputLev > alts::SILENT) {
				*out << "<" << stringArg << ">" << endl;
			}
		}
		//--------------------------------------------------------------
		//	clear-points
		//--------------------------------------------------------------
		else if (directive == "clear-points") {
			params.points.clear();
		}
		//--------------------------------------------------------------
		//	generate-points operation
		//--------------------------------------------------------------
		else if (directive == "generate-points") {
			*in >> intArg;				// number of points to generate
			params.estimatedTrueFit =
				alts::generatePoints(intArg, params.points, distrParams);
										// print summary
			if (params.outputLev >= alts::RESULTS) {
				printDistrSummary(distrParams);
			}
			dataSummarized = false;		// data set has changed
		}
		//--------------------------------------------------------------
		//	read-points operations
		//--------------------------------------------------------------
		else if (directive == "read-points") {
			*in >> stringArg;			// input file name
			readPoints(params.dim, params.points, stringArg);
			dataSummarized = false;		// data set changed
		}
		//--------------------------------------------------------------
		//	scramble-points option
		//--------------------------------------------------------------
		else if (directive == "scramble-points") {
			scramblePoints(params.points);	// scramble the point set
			dataSummarized = false;		// data set changed
		}
		//--------------------------------------------------------------
		//	seed operation
		//--------------------------------------------------------------
		else if (directive == "seed") {	// set random seed
			*in >> intArg;
			alts::setSeed(intArg);
		}
		//--------------------------------------------------------------
		//	Proces run-using
		//--------------------------------------------------------------
		else if (directive == "run-using") {
			*in >> stringArg;
			intArg = lookUp(stringArg, algoNames, alts::N_ALGS);
			if (intArg < 0) {
				cerr << "Method =  \"" << stringArg << "\"\n";
				alts::error("Illegal method", true);
			}
			params.srchParams.algorithm = (alts::Algorithm) intArg;

			params.finalize();			// finalize parameters
										// print points
			if (params.outputLev >= alts::PRINT_POINTS && !dataSummarized) {
				printPoints(params.dim, params.points);
				dataSummarized = true;
			}

			long clock0 = clock();		// start the clock
										// final result storage
			alts::LTSResult result(params.dim);
										// run it
			switch (params.srchParams.algorithm) {
				case alts::ADAPTIVE_LTS:
					if (!(slopeRangeGiven ||
						params.srchParams.slopeSampleSize >= 2)) {
						alts::error("For ALTS, either slope-range or slope-sample-size must be given", true);
					}
					result = alts::ltsSearch(params);
					break;
				case alts::FAST_LTS:
					result = alts::fastLTS(params);
					break;
				case alts::FULL_SEARCH_LTS:
					result = alts::fullSearchLTS(params);
					break;
				default:
					assert(false);
			}
										// stop the clock
			double cpuSec = double(clock() - clock0) /
							double(ALTS_CLOCKS_PER_SEC);
										// print summary
			if (params.outputLev >= alts::RESULTS) {
				printSummary(params, distrParams, result, cpuSec);
			}
		}
		//--------------------------------------------------------------
		//	Anything else - unrecognized directive
		//--------------------------------------------------------------
		else {
			cerr << "Directive: " << directive << "\n";
			alts::error("Unknown directive", true);
		}
	}
}

//----------------------------------------------------------------------
//	Print a point set
//----------------------------------------------------------------------
void printPoints(
	int							dim,
	const alts::PointSet&		points)
{
	int n = (int) points.size();		// number of points

	*out << "[data-summary:\n";
	*out << "  dimension         = " << dim << "\n";
	*out << "  data-size         = " << n << "\n";
	*out << "  (points:\n";
										// set higher precision
	out->precision(ALTS_POINT_PRECISION);
	for (int i = 0; i < n; i++) {
		const alts::Indep& indep = points[i].getIndep();
		*out << "    < ind:";
		for (int j = 1; j < dim; j++) {
			*out << " " << indep[j];
		}
		*out << " dep: " << points[i].getDep() << " >\n";
	}
	*out << "  )\n"
		 << "]" << endl;
										// default precision
	out->precision(ALTS_OUTPUT_PRECISION);
}

//------------------------------------------------------------------------
//	readPoints - read points from a given input file
//	Points are input from the given named file and appended to the point 
//	set.
//------------------------------------------------------------------------

void readPoints(
	int						d,			// dimension
	alts::PointSet&			points,		// point set (modified)
	const string&			fileName)	// file name
{
	ifstream inFile(fileName.c_str());	// open data file
	if (!inFile) {
		alts::error("Cannot open input file", true);
	}

	alts::Point tempPt(d);
	inFile >> tempPt;					// read a point
	while (!inFile.eof()) {				// successful?
		points.push_back(tempPt);		// append to point array
		inFile >> tempPt;
	}
}

//----------------------------------------------------------------------
//	scramblePoints - apply a random permutation to the point array
//
//	This is done by swapping p[i] with a random element of p[i..n-1] for 
//	i = 1,...,n-2.
//----------------------------------------------------------------------

void scramblePoints(
	alts::PointSet&	points)				// the points
{
	int n = (int) points.size();		// number of points
	for (int i = 0; i < n-1; i++) {
										// r = random from [i..n-1]
		int r = alts::randomInteger(i,n);
		if (r != i) {					// swap points[i] <--> points[r]
			alts::Point t = points[r];
			points[r] = points[i];
			points[i] = t;
		}
	}
}

//----------------------------------------------------------------------
//	printDistrSummary - print distribution summary
//----------------------------------------------------------------------

void printDistrSummary(
	const alts::DistrParams&	distrParams)// distribution parameters
{
	*out << "[distr:\n"
		 << "  type             = "
				<< distrNames[distrParams.distrib] << "\n"
		 << "  dimension        = "
				<< distrParams.dim << "\n"
		 << "  indep-range      = ["
				<< distrParams.indepRange.lo << " "
				<< distrParams.indepRange.hi << "]\n"
		 << "  dep-range        = ["
				<< distrParams.depRange.lo << " "
				<< distrParams.depRange.hi << "]\n"
		 << "  prob-inlier      = "
				<< distrParams.probInlier << "\n"
		 << "  intercept-range  = ["
				<< distrParams.interceptRange.lo << " "
				<< distrParams.interceptRange.hi << "]\n"
		 << "  slope-range      = ["
				<< distrParams.slopeRange.lo << " "
				<< distrParams.slopeRange.hi << "]\n"
		 << "  resid-std-dev    = "
				<< distrParams.residStdDev << "\n"
		 << "  indep-std-dev    = "
				<< distrParams.indepStdDev << "\n"
		 << "]" << endl;
}

//----------------------------------------------------------------------
//	Do validation
//----------------------------------------------------------------------

void doValidation(
	const alts::Params&		params,		// input parameters
	const alts::LTSResult&	result)		// the result
{
										// get cost of estimated true fit
	const alts::Plane& estTrueFit = params.estimatedTrueFit;
	alts::Scalar trueVar = alts::getLTSvariance(params.points,
				params.trimLowerValue, estTrueFit);
										// create result object
	alts::LTSResult trueFit(trueVar, estTrueFit);
										// get errors wrt to true
	alts::Scalar absErr = result.getAbsErr(trueFit);
	alts::Scalar relErr = result.getRelErr(trueFit);

	*out << "  est-true-plane    = " << trueFit.getPlane() << "\n";
	*out << "  est-true-cost     = " << trueFit.getCost() << "\n";
	*out << "  absolute-error    = " << absErr << "\n";
	*out << "  relative-error    = " << relErr << "\n";

	if (absErr > params.residAbsErr + alts::SCALAR_EPS &&
		relErr > params.residRelErr + alts::SCALAR_EPS &&
		params.srchParams.algorithm == alts::ADAPTIVE_LTS) {
		alts::error("Relative or absolute error exceeds allowable bounds", false);
	}
										// check whether true is in range
	if (params.srchParams.algorithm == alts::ADAPTIVE_LTS) {
		int d = params.dim;				// dimension
		const alts::SlopeRange& rootSlopeRange =
				params.srchParams.rootSlopeRange;
		bool inRange = true;
		for (int j = 1; j < d; j++) {
			if (estTrueFit[j] < rootSlopeRange.lo[j] ||
				estTrueFit[j] > rootSlopeRange.hi[j]) inRange = false;
		}
		if (!inRange) {
			alts::error("Adaptive-LTS slope range does not contain estimated true fit", false);
		}
	}
}

//----------------------------------------------------------------------
//	Printing execution summary
//
//	We make use of a small utility for outputting search priorities.
//----------------------------------------------------------------------

										// output priority list
std::ostream& operator<<(std::ostream& out, const alts::SrchPriorList& p)
{
	out << "( ";
	alts::SrchPriorList::const_iterator it = p.begin();
	while (it < p.end()-1) {
		out << priorityNames[*(it++)] << " ";
	}
	out << priorityNames[*it] << " )";
	return out;
}

void printSummary(
	const alts::Params&			params,		// input parameters
	const alts::DistrParams&	distrParams,// distribution parameters
	const alts::LTSResult&		result,		// the result
	double						cpuSec)		// execution time
{
	const alts::SrchParams& srchParams = params.srchParams;

	*out << "[ALTS Execution Summary:\n";
	*out << "  Params:----------------------------------------------------\n";
	*out << "  dimension         = " << params.dim << "\n";
	*out << "  output-level      = " << outLevNames[params.outputLev] << "\n";
	*out << "  data-size         = " << params.points.size() << "\n";
	*out << "  resid-abs-error   = " << params.residAbsErr << "\n";
	*out << "  resid-rel-error   = " << params.residRelErr << "\n";
	*out << "  trim-value        = " << params.trimAbsValue << " ( " << params.trimRelValue*100 << " % )\n";
	*out << "  trim-err          = " << params.trimAbsErr << " ( " << params.trimRelErr*100 << " % )\n";
	*out << "  trace-frequency   = " << params.traceFreq << "\n";
	*out << "  algorithm         = " << algoNames[srchParams.algorithm] << "\n";
	*out << "  num-pivots        = " << srchParams.numPivots << "\n";

	if (srchParams.algorithm == alts::ADAPTIVE_LTS) {
		*out << "  ALTS:-------------------------------------------------------\n";
		*out << "   max-cells        = " << srchParams.maxCells << "\n";
		*out << "   search-priority  = " << srchParams.srchPriorList << "\n";
		*out << "   slope-range      = [" << srchParams.initSlopeRange.lo
			 << " " << srchParams.initSlopeRange.hi << "]\n";
		*out << "   slope-sample-size = " << srchParams.slopeSampleSize << "\n";
		*out << "   slope-expansion  = " << srchParams.slopeExpansion << "\n";
		*out << "   root-slope-range = [" << srchParams.rootSlopeRange.lo
			 << " " << srchParams.rootSlopeRange.hi << "]\n";
		*out << "   split-method     = " << splitNames[srchParams.splitMethod] << "\n";
	}
	else if (srchParams.algorithm == alts::FAST_LTS) {
		*out << "  FLTS:------------------------------------------------------\n";
		*out << "   num-elem-fits    = " << srchParams.numElemFits << "\n";
		*out << "   num-best         = " << srchParams.numBest << "\n";
	}
	else { // (srchParams.algorithm == FULL_SEARCH_LTS)
		*out << "  Full:------------------------------------------------------\n";
		*out << "   sample-size      = " << srchParams.fullSampleSize << "\n";
		*out << "   max-slabs        = " << srchParams.fullMaxSlabs << "\n";
	}

	*out << "  Results:---------------------------------------------------\n";
	*out << "  lts-plane         = " << result.getPlane() << "\n";
	*out << "  lts-cost          = " << result.getCost() << "\n";
	if (params.doValidation) {			// validate results
		doValidation(params, result);
	}
	*out << "  found-at-time     = " << result.getTimeStamp() << "\n";
	*out << "  exec-time         = " << cpuSec << " seconds";
	#ifdef PERF
		*out << " (biased by performance measurements)";
	#endif
	*out << "\n";
	*out << "  -----------------------------------------------------------\n\n";
										// output performance stats
	if (params.outputLev >= alts::PERF_STATS) {
		#ifdef PERF
										// kludge! stats only for ALTS
			if (srchParams.algorithm == alts::ADAPTIVE_LTS) {
				alts::printStats(*out, (params.outputLev >= alts::LEVEL_STATS));
			}
		#else
			*out << "  (Performance statistics unavailable.)\n";
		#endif
	}
	*out << "]" << endl;
}
